#include <abort.h>
#include <malloc.h>
#include <stream.h>
#include <ultra64.h>
#include <sys/region.h>
#include <math/vector.h>
#include <math/math.h>
#include <sys/sys.h>
#include <vertex.h>
#include <node/lux.h>
#include <node/xformer.h>
#include <types.h>
#include "dltrans.h"

#define VTXBUFFER_MAXELEMENTS	64
#define IDXBUFFER_MAXELEMENTS	64
#define MAX_LIGHTS 8
#define MAX_NODES 64
#define	MAX_MATERIALS 128

static Light rsp_light[MAX_MATERIALS][MAX_LIGHTS];
static Lights0 rsp_ambient_light = gdSPDefLights0( 0x00, 0x00, 0x40 );

static Material material[MAX_MATERIALS];
static const Lux* lux[MAX_LIGHTS];
static const Vector3f* lux_pos[MAX_LIGHTS];
static Matrix* cpu_mtx[MAX_NODES];

static Vp vp = {
	{
		{(BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ, 0},
		{(BUFFER_W/2)*4, (BUFFER_H/2)*4, 0, 0}
	}
};

static Gfx rspinit_dl[] = {
	gsSPViewport(&vp),
	gsSPClipRatio( FRUSTRATIO_2 ),
	gsSPEndDisplayList()
};

static Gfx rdpinit_dl[] = {
	gsDPSetCycleType(G_CYC_1CYCLE),
	gsDPPipelineMode(G_PM_NPRIMITIVE),
	gsDPSetScissor(G_SC_NON_INTERLACE, 0, 0, BUFFER_W, BUFFER_H),
	gsDPSetTextureLOD(G_TL_TILE),
	gsDPSetTextureLUT(G_TT_NONE),
	gsDPSetTextureDetail(G_TD_CLAMP),
	gsDPSetTexturePersp(G_TP_PERSP),
	gsDPSetTextureFilter(G_TF_BILERP),
	gsDPSetTextureConvert(G_TC_FILT),
	gsDPSetCombineMode(G_CC_SHADE, G_CC_SHADE),
	gsDPSetCombineKey(G_CK_NONE),
	gsDPSetAlphaCompare(G_AC_NONE),
	gsDPSetRenderMode(G_RM_OPA_SURF, G_RM_OPA_SURF2),
	gsDPSetColorDither(G_CD_DISABLE),
	gsDPPipeSync(),
	gsSPEndDisplayList()
};

void init_DLTrans( DLTrans* dltrans ) {
	init_Vec( &dltrans->rspMatrices,
		sizeof(Mtx),
		ALIGN_8B
	);
	resize_Vec( &dltrans->rspMatrices, MAX_NODES );

	init_Vec( &dltrans->vtxBuffer,
		sizeof(Vtx),
		ALIGN_8B
	);
	resize_Vec( &dltrans->vtxBuffer,
		VTXBUFFER_MAXELEMENTS
	);
	dltrans->idxBuffer = malloc(
		sizeof(*dltrans->idxBuffer)*IDXBUFFER_MAXELEMENTS
	);
	dltrans->scale = 1.0f;
	setColor_Material( &dltrans->material, 1.0f, 1.0f, 1.0f );
}

void destroy_DLTrans( DLTrans* dltrans ) {
	free( dltrans->idxBuffer );
	destroy_Vec( &dltrans->vtxBuffer );
}

void process_DLTrans( DLTrans* dltrans, const DList* dl_in, NativeDL* dl_out ) {
	reset_Vec( &dltrans->rspMatrices );

	NativeDL* rsp_dl = dl_out;
	uint16_t current_mtx_i = 0;
	uint16_t current_lux_i = 0;
	uint16_t current_mtl_i = 0;

	Stream stream;
	init_Stream( &stream,
		getData_DList( dl_in ),
		getSize_DList( dl_in )
	);

	name_t name;
	do {
		name = *( name_t* )getData_Stream( &stream );
		switch( name ) {
			case Start: {
				( void* )extractElement_Stream( &stream,
					sizeof( Command_Start )
				);

				addCommand_NativeDL( rsp_dl,
					gsSPSegment( 0x0, 0x0 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPSegment( 0x8, 0x0 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( rspinit_dl )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( rdpinit_dl )
				);
				break;
			}
			case SetVertexSource: {
				const Command_SetVertexSource* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetVertexSource )
					);
				dltrans->vtxFormat = cmd->format;
				dltrans->vtxSource = cmd->data;
				break;
			}
			case SetIndexSource: {
				const Command_SetIndexSource* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetIndexSource )
					);
				dltrans->idxSource = cmd->data;
				break;
			}
			case SetVertexScale: {
				const Command_SetVertexScale* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetVertexScale )
					);
				dltrans->scale = cmd->scale;
				break;
			}
			case LoadVertexBuffer: {
				const Command_LoadVertexBuffer* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadVertexBuffer )
					);

				uint32_t n = cmd->nvts;
				if( n + cmd->dst_ofs > VTXBUFFER_MAXELEMENTS ) {
					abort();
				}

				Vertex* vtxSource = ( Vertex* )dltrans->vtxSource + cmd->src_ofs;
				Vtx* vtx =
					getElement_Vec( &dltrans->vtxBuffer,
						cmd->dst_ofs
					);
				while( n-- ) {
					vtx->v.ob[0] = vtxSource->position.x * dltrans->scale;
					vtx->v.ob[1] = vtxSource->position.y * dltrans->scale;
					vtx->v.ob[2] = vtxSource->position.z * dltrans->scale;
					vtx->v.tc[0] = vtxSource->texcoords.x * 32 * getWidth_Texture( dltrans->tex );
					vtx->v.tc[1] = vtxSource->texcoords.y * 32 * getHeight_Texture( dltrans->tex );
					vtx->v.cn[0] = vtx->v.cn[1] = vtx->v.cn[2] = vtx->v.cn[3] = 0xFF;

					vtx++;
					vtxSource++;
				}

				break;
			}
			case LoadIndexBuffer: {
				const Command_LoadIndexBuffer* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadIndexBuffer )
					);

				uint32_t n = cmd->nids;
				if( n > IDXBUFFER_MAXELEMENTS ) {
					abort();
				}

				uint32_t* idxSource = ( uint32_t* )dltrans->idxSource + cmd->src_ofs;
				uint32_t* idx = dltrans->idxBuffer + cmd->dst_ofs;
				while( n-- ) {
					*idx++ = *idxSource++;
				}

				break;
			}
			case DrawPrimitives: {
				const Command_DrawPrimitives* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_DrawPrimitives )
					);

				Vtx* vtx =
					getElement_Vec( &dltrans->vtxBuffer, 0 );
				uint32_t* idx = dltrans->idxBuffer;
				if( PrimTriangle == cmd->type ) {
					for( uint32_t i=0; i<cmd->num; i++ ) {
						addCommand_NativeDL( rsp_dl,
							gsSPVertex( &vtx[*idx++], 1, 0 )
						);
						addCommand_NativeDL( rsp_dl,
							gsSPVertex( &vtx[*idx++], 1, 1 )
						);
						addCommand_NativeDL( rsp_dl,
							gsSPVertex( &vtx[*idx++], 1, 2 )
						);
						addCommand_NativeDL( rsp_dl,
							gsSP1Triangle( 0, 1, 2, 0 )
						);
					}
				}
				else {
					abort();
				}

				break;
			}
			case SetBuffer: {
				const Command_SetBuffer* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetBuffer )
					);
				Surface* surface = NULL;

				if( ColorBuffer == cmd->buftype ) {
					surface = &dltrans->surfColor;
					addCommand_NativeDL( rsp_dl,
						gsDPSetColorImage(
							G_IM_FMT_RGBA,
							G_IM_SIZ_16b,
							cmd->width,
							cmd->buf
						)
					);
				} else if( ZBuffer == cmd->buftype ) {
					surface = &dltrans->surfZ;
					addCommand_NativeDL( rsp_dl,
						gsDPSetDepthImage( cmd->buf )
					);
				} else {
					abort();
				}

				surface->pixfmt = cmd->format;
				surface->data = cmd->buf;
				surface->width = cmd->width;
				surface->height = cmd->height;

				break;
			}
			case ClearBuffers: {
				const Command_ClearBuffers* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_ClearBuffers )
					);
				uint32_t clrval;
				Surface* surface = NULL;

				if( cmd -> types & ColorBuffer ) {
					clrval = GPACK_RGBA5551(
						( uint8_t )( 255 * dltrans->clearColor.x ),
						( uint8_t )( 255 * dltrans->clearColor.y ),
						( uint8_t )( 255 * dltrans->clearColor.z ),
						1
					);
					clrval = clrval << 16 | clrval;
					surface = &dltrans->surfColor;

					addCommand_NativeDL( rsp_dl,
						gsDPSetColorImage(
							G_IM_FMT_RGBA,
							G_IM_SIZ_16b,
							surface->width,
							surface->data
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetCycleType( G_CYC_FILL )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_NOOP,
							G_RM_NOOP2
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetFillColor( clrval )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPFillRectangle(
							0,
							0,
							surface->width - 1,
							surface->height - 1
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPPipeSync()
					);
				}
				if( cmd -> types & ZBuffer ) {
					clrval = GPACK_ZDZ(
						( uint16_t )( G_MAXFBZ * dltrans -> clearZVal ),
						0
					);
					clrval = clrval << 16 | clrval;
					surface = &dltrans -> surfZ;

					addCommand_NativeDL( rsp_dl,
						gsDPSetColorImage(
							G_IM_FMT_RGBA,
							G_IM_SIZ_16b,
							surface->width,
							surface->data
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetCycleType( G_CYC_FILL )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_NOOP,
							G_RM_NOOP2
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetFillColor( clrval )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPFillRectangle(
							0,
							0,
							surface->width - 1,
							surface->height - 1
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPPipeSync()
					);
				}

				break;
			}
			case SetClearColor: {
				const Command_SetClearColor* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetClearColor )
					);

				init_Vector4f( &dltrans->clearColor, cmd->r, cmd->g, cmd->b, cmd->a );

				break;
			}
			case SetClearZValue: {
				const Command_SetClearZValue* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetClearZValue )
					);

				dltrans->clearZVal = cmd->val;

				break;
			}
			case LoadTexture: {
				const Command_LoadTexture* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadTexture )
					);

				dltrans->tex = cmd->texture;
				const Texture* texture = cmd->texture;
				if( TEXTURE_FMT_RGBA32 == texture->fmt ) {
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTextureBlock(
							texture->pixels,
							G_IM_FMT_RGBA,
							G_IM_SIZ_32b,
							texture->width,
							texture->height,
							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2uint( texture->width ),
							log2uint( texture->height ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
					);
					addCommand_NativeDL(rsp_dl, gsDPSetTextureLUT(G_TT_NONE));
				} else if( TEXTURE_FMT_RGBA16 == texture->fmt ) {
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTextureBlock(
							texture->pixels,
							G_IM_FMT_RGBA,
							G_IM_SIZ_16b,
							texture->width,
							texture->height,
							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2uint( texture->width ),
							log2uint( texture->height ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetTextureLUT( G_TT_NONE )
					);
				} else if( TEXTURE_FMT_CI4 == texture->fmt ) {
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTextureBlock_4b(
							texture->pixels,
							G_IM_FMT_CI,
							texture->width,
							texture->height,
							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2uint( texture->width ),
							log2uint( texture->height ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTLUT_pal16(
							0,
							texture->palette
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetTextureLUT( G_TT_RGBA16 )
					);
				} else {
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTextureBlock(
							texture->pixels,
							G_IM_FMT_RGBA,
							G_IM_SIZ_32b,
							texture->width,
							texture->height,
							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2uint( texture->width ),
							log2uint( texture->height ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
					);
					addCommand_NativeDL(rsp_dl, gsDPSetTextureLUT(G_TT_NONE));
					//abort( );
				}
				break;
			}
			case Jump: {	// beware: allocation on every invocation!
				abort();
				const Command_Jump* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_Jump )
					);
				NativeDL* child_dl = new_NativeDL();

				process_DLTrans( dltrans, cmd->dl, child_dl );
				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( child_dl )
				);

				break;
			}
			case End: {
				(void*)extractElement_Stream( &stream,
					sizeof( Command_End )
				);

				addCommand_NativeDL( rsp_dl,
					gsSPEndDisplayList( )
				);
				break;
			}
			case SetTextureScale: {
				const Command_SetTextureScale* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetTextureScale )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPTexture(
						( s32 )( 0x8000 * cmd->s ),
						( s32 )( 0x8000 * cmd->t ),
						0,
						G_TX_RENDERTILE,
						G_ON
					)
				);

				// temporary stuff
				addCommand_NativeDL( rsp_dl,
					gsDPPipeSync( )
				);
				addCommand_NativeDL( rsp_dl,
					gsDPSetCycleType( G_CYC_1CYCLE )
				);
				//addCommand_NativeDL(rsp_dl, gsDPSetRenderMode(G_RM_ZB_OPA_SURF, G_RM_ZB_OPA_SURF2));

				break;
			}
			case LoadMatrix: {
				const Command_LoadMatrix* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadMatrix )
					);

				if( current_mtx_i >= MAX_NODES ) {
					abort();
				}

				cpu_mtx[current_mtx_i] = cmd->mtx;

				Mtx* mtx_native = getElement_Vec( &dltrans->rspMatrices, current_mtx_i++ );
				guMtxF2L(
					cmd->mtx->m,
					mtx_native
				);

				if( MATRIX_PROJECTION == cmd->type ) {
					addCommand_NativeDL( rsp_dl,
						gsSPMatrix(
							mtx_native,
							G_MTX_PROJECTION | G_MTX_LOAD | G_MTX_NOPUSH
						)
					);
				} else if( MATRIX_MODELVIEW == cmd -> type ) {
					addCommand_NativeDL( rsp_dl,
						gsSPMatrix(
							mtx_native,
							G_MTX_MODELVIEW | G_MTX_LOAD | G_MTX_NOPUSH
						)
					);
				} else {
					abort( );
				}
				break;
			}
			case LoadLight: {
				const Command_LoadLight* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadLight )
					);

				uint16_t idx = cmd->idx;
				if( idx > MAX_LIGHTS ) {
					abort();
				}

				/* color */
				lux[current_lux_i] = cmd->light;
				lux_pos[current_lux_i] = cmd->pos;
				++current_lux_i;

				break;
			}
			case Sync: {
				( void* )extractElement_Stream( &stream,
					sizeof( Command_Sync )
				);

				addCommand_NativeDL( rsp_dl,
					gsDPFullSync()
				);
				break;
			}
			case PipeSync: {
				( void* )extractElement_Stream( &stream,
					sizeof( Command_PipeSync )
				);

				addCommand_NativeDL( rsp_dl,
					gsDPPipeSync()
				);
				break;
			}
			case JumpToTranslated: {
				const Command_JumpToTranslated* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_JumpToTranslated )
					);

				uint16_t mtx_i = current_mtx_i - 1;
				if( 0 == current_mtx_i ) {
					mtx_i = 0;
				}
				uint16_t mtl_i = current_mtl_i - 1;
				if( 0 == current_mtl_i ) {
					mtl_i = 0;
				}

				Vector3f center, dir;
				Matrix* mtx;
				for( uint16_t i=0; i<current_lux_i; ++i ) {
					mtx = cpu_mtx[mtx_i];
					init_Vector3f(
						&center,
						MATRIX_GET(*mtx, 3, 0),
						MATRIX_GET(*mtx, 3, 1),
						MATRIX_GET(*mtx, 3, 2)
					);
					sub_Vector3f( &dir, &center, lux_pos[i] );
					normalize_Vector3f( &dir, &dir );

					Light_t* light = &rsp_light[mtl_i][i].l;
					const Vector3f* mtl_color = &material[mtl_i].color;
					/* multiply lights by material colors before rendering */
					light->col[0]
						= light->colc[0]
						= 255 * lux[i]->color.x * mtl_color->x;

					light->col[1]
						= light->colc[1]
						= 255 * lux[i]->color.y * mtl_color->y;

					light->col[2]
						= light->colc[2]
						= 255 * lux[i]->color.z * mtl_color->z;

					/* direction */
					light->dir[0] = -127*dir.x;
					light->dir[1] = -127*dir.y;
					light->dir[2] = -127*dir.z;

					addCommand_NativeDL( rsp_dl,
						gsSPLight( &rsp_light[mtl_i][i], i + 1 )
					);
				}
				addCommand_NativeDL( rsp_dl,
					gsSPLight( &rsp_ambient_light, current_lux_i + 1 )
				);

				s32 num_lights;
				switch( current_lux_i ) {
					case 0:
						num_lights = NUMLIGHTS_0;
						break;
					case 1:
						num_lights = NUMLIGHTS_1;
						break;
					case 2:
						num_lights = NUMLIGHTS_2;
						break;
					case 3:
						num_lights = NUMLIGHTS_3;
						break;
					case 4:
						num_lights = NUMLIGHTS_4;
						break;
					case 5:
						num_lights = NUMLIGHTS_5;
						break;
					case 6:
						num_lights = NUMLIGHTS_6;
						break;
					case 7:
						num_lights = NUMLIGHTS_7;
						break;
					default:
						num_lights = NUMLIGHTS_7;
						break;
				}
				addCommand_NativeDL( rsp_dl,
					gsSPNumLights( num_lights )
				);

				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( ( Gfx* ) cmd -> dl )
				);
				break;
			}
			case SetCombineMode: {
				const Command_SetCombineMode* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetCombineMode )
					);

				if( CM_SHADE == cmd -> mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetCombineMode(
							G_CC_SHADE,
							G_CC_SHADE
						)
					);
				} else if( CM_DECAL == cmd -> mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetCombineMode(
							G_CC_DECALRGB,
							G_CC_DECALRGB
						)
					);
				} else {
					abort( );
				}
				break;
			}
			case SetRenderMode: {
				const Command_SetRenderMode* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetRenderMode )
					);

				if( RM_ZB_OPAQUE == cmd -> mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_ZB_OPA_SURF,
							G_RM_ZB_OPA_SURF2
						)
					);
				} else if( RM_OPAQUE == cmd -> mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_OPA_SURF,
							G_RM_OPA_SURF2
						)
					);
				} else {
					abort( );
				}
				break;
			}
			case NormalizePerspective: {
				const Command_NormalizePerspective* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_NormalizePerspective )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPPerspNormalize( ( u16 ) cmd -> scale)
				);
				break;
			}
			case SetLookAt: {
				const Command_SetLookAt* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetLookAt )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPLookAt( cmd -> lookat )
				);
				break;
			}
			case SetGeometryMode: {
				const Command_SetGeometryMode* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetGeometryMode )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPSetGeometryMode( cmd -> gmode )
				);
				break;
			}
			case ClearGeometryMode: {
				const Command_ClearGeometryMode* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_ClearGeometryMode )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPClearGeometryMode( cmd -> gmode )
				);
				break;
			}
			case SetViewport: {
				const Command_SetViewport* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetViewport )
					);

				vp.vp.vscale[0] = (cmd->vp->width/2)*4;
				vp.vp.vscale[1] = (cmd->vp->height/2)*4;
				vp.vp.vscale[2] = G_MAXZ;
				vp.vp.vscale[3] = 0;

				vp.vp.vtrans[0] = (cmd->vp->width/2)*4;
				vp.vp.vtrans[1] = (cmd->vp->height/2)*4;
				vp.vp.vtrans[2] = 0;
				vp.vp.vtrans[3] = 0;

				addCommand_NativeDL( rsp_dl,
					gsSPViewport( &vp )
				);
				break;
			}

			case SetMaterial: {
				const Command_SetMaterial* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetMaterial )
					);

				if( current_mtl_i > MAX_MATERIALS ) {
					abort();
				}
				copy_Material( &material[current_mtl_i], cmd->mtl );
				++current_mtl_i;

				break;
			}
			
			default: {
				abort( );
				break;
			}
		}
	} while( End != name );

	destroy_Stream( &stream );
}

