#include <abort.h>
#include <float.h>
#include <sys/region.h>
#include <node/xformer_funcs.h>
#include "engine.h"

#define MAX_NODES	64

static bool mesh_to_vpairs_equ( void* key1, void* key2 ) {
	return key1 == key2;
}

void init_Engine( Engine* engine ) {
	init_Vec( &engine->lights, sizeof(Lux*), ALIGN_DEFAULT );
	init_Vec( &engine->cpuMatrixes, sizeof(Matrix), ALIGN_4B );
	resize_Vec( &engine->cpuMatrixes, MAX_NODES );
	hash_Construct(
		&engine->mesh_to_vpairs,
		sizeof(Mesh*),
		sizeof(Vec*),
		mesh_to_vpairs_equ
	);
	init_DList( &engine->dl );

	init_NativeDL( &engine->rsp_cfb );
	init_NativeDL( &engine->rsp_zb );

	//
	init_Vec(
		&engine->xformers,
		sizeof(Xformer),
		ALIGN_DEFAULT			// how to get a structure's alignment?
	);
	resize_Vec( &engine->xformers, 30 );	// don't mess with the vec from now - we don't want any reallocations to occur

	init_Vec(
		&engine->meshes,
		sizeof(Mesh),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->meshes, 30 );

	init_Vec(
		&engine->billboards,
		sizeof(Billboard),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->billboards, 30 );

	init_Vec(
		&engine->cameras,
		sizeof(Camera),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->cameras, 30 );

	init_Vec(
		&engine->luxes,
		sizeof(Lux),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->luxes, 30 );
}

inline void setViewport_Engine( Engine* engine, Viewport* viewport ) {
	engine->viewport = viewport;
}

inline void setCamera_Engine( Engine* engine, Camera_Xformer* camera_xf ) {
	engine->camera_xf = camera_xf;
}

inline void setRenderTarget_Engine( Engine* engine, const Surface* renderTarget ) {
	engine->renderTarget = renderTarget;
}

inline void setZBuffer_Engine( Engine* engine, const Surface* zbuffer ) {
	engine->zBuffer = zbuffer;
}

inline Viewport* getViewport_Engine( const Engine* engine ) {
	return engine->viewport;
}

inline Camera_Xformer* getCamera_Engine( const Engine* engine ) {
	return engine->camera_xf;
}

inline const Surface* getRenderTarget_Engine( const Engine* engine ) {
	return engine->renderTarget;
}

inline const Surface* getZBuffer_Engine( const Engine* engine ) {
	return engine->zBuffer;
}

inline void addLight_Engine( Engine* engine, const Lux* light ) {
	pushBack_Vec( &engine->lights, &light );
}

void removeLight_Engine( Engine* engine, const Lux* light ) {
	if(engine) {}
	if(light) {}
}

static void handleMesh( DList* dlist, const Mesh_Xformer* mesh_xf ) {
	pushLoadMatrix_DList(
		dlist,
		getResultMatrix_Xformer( mesh_xf->xf ),
		MATRIX_MODELVIEW
	);

	const Gfx* meshdl;
	const Material* mat;
	const Pair* pair;
	for( uint32_t i=0, n=getLength_Vec( &mesh_xf->mesh->vpairs ); i<n; ++i ) {
		pair = getConstElement_Vec( &mesh_xf->mesh->vpairs, i );
		meshdl = pair->gfx;
		mat = pair->mat;

		if( NULL == mat || NULL == meshdl )
			abort();

		pushSetMaterial_DList( dlist, mat );
		//mat->tex = NULL;
		if( mat->tex ) {
			pushSetCombineMode_DList( dlist, CM_DECAL );
			pushLoadTexture_DList( dlist, mat->tex );
		} else {
			pushSetCombineMode_DList( dlist, CM_SHADE );
#if 0
			dlistSetTextureScale(&engine->dl, 0.125f, 0.125f);
			dlistSetRenderMode(&engine->dl, RM_DECAL);
			dlistSetGeometryMode(&engine->dl, GM_LIGHTING | GM_TEXGEN);
			const Coordsys* camcs = tnodeGetCoordsys((TNode*)engineGetCamera(engine));
			const Coordsys* meshcs = tnodeGetCoordsys((TNode*)mesh);
			static Matrix mreflect;
			static LookAt lat;
			guLookAtReflectF(mreflect, &lat,
				camcs->origin.x, camcs->origin.y, camcs->origin.z,
				meshcs->n.x, meshcs->n.y, meshcs->n.z,
				0.0f, 1.0f, 0.0f);
			dlistSetLookAt(&engine->dl, &lat);
#endif
		}

		pushJumpToTranslated_DList( dlist, meshdl );
	}
}

static void handleBillboard( DList* dlist, const Billboard_Xformer* bb_xf ) {
	handleMesh( dlist, (Mesh_Xformer*)bb_xf );
}

void sceneTaskF_Engine( Engine* engine, TreeNode* scene ) {
	const Viewport* vp = getViewport_Engine( engine );
	const Camera_Xformer* cam_xf = getCamera_Engine( engine );
	float perspNorm;
	perspective_Matrix(
		&engine->projMtxF,
		&perspNorm,
		getFOVY_Camera( cam_xf->camera ),
		getWidth_Viewport( vp ) / getHeight_Viewport( vp ),
		getNear_Camera( cam_xf->camera ),
		getFar_Camera( cam_xf->camera )
	);
	pushNormalizePerspective_DList( &engine->dl,
		perspNorm
	);

	const Matrix* mtx = getMatrix_Xformer( cam_xf->xf );
	Matrix inv;
	invert_Matrix( mtx, &inv );
	multiply_Matrix( &engine->projMtxF, &inv, &engine->projMtxF );
	pushLoadMatrix_DList(
		&engine->dl,
		&engine->projMtxF,
		MATRIX_PROJECTION
	);

	pushSetGeometryMode_DList(
		&engine->dl,
		GM_CULLBACK |
		GM_ZBUFFER |
		GM_LIGHTING |
		GM_SHADE |
		GM_GOURAUD
	);
	pushSetTextureScale_DList( &engine->dl, 1.0f, 1.0f );
	pushPipeSync_DList( &engine->dl );
	pushSetRenderMode_DList( &engine->dl, RM_ZB_OPAQUE );

	static TreeNodeIterator ni;
	static bool init = true;
	if( init ) {
		init_TreeNodeIterator( &ni );
		init = false;
	}
	reset_TreeNodeIterator( &ni, scene );

	TreeNode* tn;
	Matrix* cpuMtx = getElement_Vec( &engine->cpuMatrixes, 0 );
	// 1st pass - calculate world coordintes of nodes
	while( (tn = inc_TreeNodeIterator( &ni )) ) {
		Xformer* parent_xf = NULL;
		if( getParent_TreeNode( tn ) ) {
			parent_xf = ( (Raw_Xformer*)getData_TreeNode( getParent_TreeNode( tn ) ) )->xf;
		}
		finish_Xformer(
			( (Raw_Xformer*)getData_TreeNode( tn ) )->xf,
			parent_xf,
			cpuMtx
		);
		++cpuMtx;
	}
	reset_TreeNodeIterator( &ni, scene );

	// 2nd pass - calculate lighting
	while( (tn = inc_TreeNodeIterator( &ni )) ) {
		const Raw_Xformer* raw_xf = getData_TreeNode( tn );

		switch( raw_xf->tag ) {
			case OBJECT_LUX: {
				pushLoadLight_DList(
					&engine->dl,
					( (Lux_Xformer*)raw_xf )->lux,
					&getCoordsys_Xformer( raw_xf->xf )->origin,
					//&getCoordsys_Xformer( raw_xf->xf )->n,
					0
				);
				break;
			}

			default:
				break;
		}
	}
	reset_TreeNodeIterator( &ni, scene );

	// 3rd pass - apply additional transformations and build display list
	while( (tn = inc_TreeNodeIterator( &ni )) ) {
		Raw_Xformer* raw_xf = getData_TreeNode( tn );
		void* data = NULL;
		switch( raw_xf->tag ) {
			case OBJECT_BILLBOARD: {
				data = cam_xf->xf;
				break;
			}
			default:
				break;
		}

		if( raw_xf->xf->processor ) {
			raw_xf->xf->processor( (struct Xformer*)raw_xf->xf, data );
		}

		reset_Xformer( raw_xf->xf );

		switch( raw_xf->tag ) {
			case OBJECT_BILLBOARD:
				handleBillboard( &engine->dl, (Billboard_Xformer*)raw_xf );
				break;
			case OBJECT_MESH: {
				handleMesh( &engine->dl, (Mesh_Xformer*)raw_xf );
				break;
			}

			default:
				break;
		}
	}
}

// This uses deprecated code
void overlayTaskF_Engine( Engine* engine, void* arg ) {
	return;
	Object2DTNode* sprite = arg;

	static Vertex rect_vts[] = {
		{{-0.5, -0.5, 0}, {0, 1}},
		{{+0.5, -0.5, 0}, {1, 1}},
		{{+0.5, +0.5, 0}, {1, 0}},
		{{-0.5, +0.5, 0}, {0, 0}},
	};
	static const uint32_t rect_ids[] = {
		0, 1, 3,
		1, 2, 3,
	};

	//
	static NodeIterator ni;
	static bool init = true;
	if( init ) {
		nodeIteratorConstruct( &ni );
		init = false;
	}
	nodeIteratorReset( &ni, (Node*)sprite );

	TNode* tnode;
	Matrix* cpuMtx = getElement_Vec( &engine->cpuMatrixes, 0 );
	// 1st pass - calculate world coordintes of nodes
	while( (tnode = (TNode*)nodeIteratorInc( &ni )) ) {
		tnodeProcess( tnode, cpuMtx );
		++cpuMtx;
	}
	nodeIteratorReset( &ni, (Node*)sprite );
	// 2nd pass - build display list
#define VSCALE	(1024.0f)
	static Matrix identity;
	static Matrix mtx;
	identity_Matrix( &identity );
	scale_Matrix( &mtx,
		1.0f/VSCALE,
		1.0f/VSCALE,
		1.0f
	);
	pushClearGeometryMode_DList( &engine->dl, GM_ZBUFFER );
	pushPipeSync_DList( &engine->dl );
	pushSetRenderMode_DList( &engine->dl, RM_OPAQUE );
	pushSetVertexSource_DList( &engine->dl, &rect_vts, 0 );
	pushSetIndexSource_DList( &engine->dl, &rect_ids );
	pushLoadMatrix_DList( &engine->dl, &identity, MATRIX_PROJECTION );
	pushSetVertexScale_DList( &engine->dl, VSCALE );
	pushLoadVertexBuffer_DList( &engine->dl, 0, 0, 4 );
	pushLoadIndexBuffer_DList( &engine->dl, 0, 0, 6 );
	pushSetTextureScale_DList( &engine->dl, 1.0f, 1.0f );
	pushSetCombineMode_DList( &engine->dl, CM_DECAL );

	while( (tnode = (TNode*)nodeIteratorInc( &ni )) ) {
		if( tnode->Process ) {
			tnode->Process( tnode, NULL );
		}

		tnodeReset( tnode );

		multiply_Matrix(
			tnodeGetTransformedMatrix( tnode ),
			&mtx,
			tnodeGetTransformedMatrix( tnode )
		);
		pushLoadMatrix_DList(
			&engine->dl,
			tnodeGetTransformedMatrix( tnode ),
			MATRIX_MODELVIEW
		);

		pushLoadTexture_DList(
			&engine->dl,
			( (const Pair*)getElement_Vec( &sprite->object2d.mesh.vpairs, 0 ) )->mat->tex
		);
		pushDrawPrimitives_DList( &engine->dl, PrimTriangle, 2 );
#undef VSCALE
	}
}

void clearColorbufF_Engine( Engine* engine, void* arg ) {
	const Vector3f* color = arg;
	pushSetClearColor_DList( &engine->dl,
		color->x,
		color->y,
		color->z,
		1.0f
	);
	pushClearBuffers_DList( &engine->dl, ColorBuffer );
}

void clearZbufF_Engine( Engine* engine, void* arg ) {
	const float* zval = arg;
	pushSetClearZValue_DList( &engine->dl, *zval );
	pushClearBuffers_DList( &engine->dl, ZBuffer );
}

void processTask_Engine( Engine* engine, const Task* task ) {
	const Surface* cbuf = getRenderTarget_Engine( engine );
	const Surface* zbuf = getZBuffer_Engine( engine );

	reset_DList( &engine->dl );
	pushStart_DList( &engine->dl );
	pushSetViewport_DList( &engine->dl, getViewport_Engine( engine ) );
	pushSetBuffer_DList(
		&engine->dl,
		ColorBuffer,
		0,
		getWidth_Surface( cbuf ),
		getHeight_Surface( cbuf ),
		getData_Surface( cbuf )
	);
	pushSetBuffer_DList(
		&engine->dl,
		ZBuffer,
		0,
		getWidth_Surface( zbuf ),
		getHeight_Surface( zbuf ),
		getData_Surface( getZBuffer_Engine( engine ) )
	);

	task->f( engine, task->arg );

	pushSync_DList( &engine->dl );
	pushEnd_DList( &engine->dl );
}

Xformer* createXformer_Engine( Engine* self, ProcessFunc_Xformer processor ) {
	Xformer xf;
	init_Xformer( &xf, processor );
	pushBack_Vec( &self->xformers, &xf );

	return getElement_Vec(
		&self->xformers,
		getLength_Vec( &self->xformers ) - 1
	);
}

Raw_Xformer createRaw_Engine( Engine* self ) {
	Raw_Xformer raw_xf;

	raw_xf.tag = OBJECT_RAW;
	raw_xf.xf = createXformer_Engine( self, NULL );

	return raw_xf;
}

Mesh_Xformer createMesh_Engine( Engine* self, const Pair* pairs ) {
	Mesh_Xformer mesh_xf;

	Mesh mesh;
	init_Mesh( &mesh, pairs );
	pushBack_Vec( &self->meshes, &mesh );
	mesh_xf.tag = OBJECT_MESH;
	mesh_xf.mesh = getElement_Vec(
		&self->meshes,
		getLength_Vec( &self->meshes ) - 1
	);

	mesh_xf.xf = createXformer_Engine( self, NULL );

	return mesh_xf;
}

Billboard_Xformer createBillboard_Engine( Engine* self, const Pair* pairs ) {
	Billboard_Xformer bb_xf;

	Billboard bb;
	init_Billboard( &bb, pairs );
	pushBack_Vec( &self->billboards, &bb );

	bb_xf.tag = OBJECT_BILLBOARD;
	bb_xf.billboard = getElement_Vec(
		&self->billboards,
		getLength_Vec( &self->billboards ) - 1
	);

	bb_xf.xf = createXformer_Engine( self, processBillboard_Xformer );

	return bb_xf;
}

Camera_Xformer createCamera_Engine( Engine* self ) {
	Camera_Xformer cam_xf;

	Camera cam;
	init_Camera( &cam );
	pushBack_Vec( &self->cameras, &cam );

	cam_xf.tag = OBJECT_CAMERA;
	cam_xf.camera = getElement_Vec(
		&self->cameras,
		getLength_Vec( &self->cameras ) - 1
	);

	cam_xf.xf = createXformer_Engine( self, processCamera_Xformer );

	return cam_xf;
}

Lux_Xformer createLux_Engine( Engine* self ) {
	Lux_Xformer lux_xf;

	Lux lux;
	init_Lux( &lux );
	pushBack_Vec( &self->luxes, &lux );
	lux_xf.lux = getElement_Vec(
		&self->luxes,
		getLength_Vec( &self->luxes ) - 1
	);

	lux_xf.tag = OBJECT_LUX;
	lux_xf.xf = createXformer_Engine( self, NULL );

	return lux_xf;
}

