#ifndef DLIST_H
#define DLIST_H

#include <math/matrix.h>
#include <math/vector.h>
#include <node/lux.h>
#include <material.h>
#include <vertex.h>
#include <viewport.h>
#include <texture.h>
#include <types.h>

typedef uintmax_t	name_t;	// i wonder if this always enforces proper alignment

typedef struct {
	void*	data;
	void*	current;
	size_t	size;
} DList;

typedef Vertex_pos3f_tc2f Vertex;

#if 0
typedef struct {
	Vector3f	position;
	Vector2f	texcoords;
} Vertex;
#endif
typedef enum {
	PrimTriangle
} DListPrimitiveType;

typedef enum {
	ColorBuffer = 0x01,
	ZBuffer = 0x02,
} DListBufferType;

typedef enum {
	MATRIX_PROJECTION,
	MATRIX_MODELVIEW,
} MatrixType;

typedef enum {
	CM_SHADE,
	CM_DECAL,
} CombineMode;

typedef enum {
	RM_ZB_OPAQUE,
	RM_OPAQUE,
} RenderMode;

typedef enum {
	GM_LIGHTING = G_LIGHTING,
	GM_TEXGEN = G_TEXTURE_GEN,
	GM_ZBUFFER = G_ZBUFFER,
	GM_CULLBACK = G_CULL_BACK,
	GM_SHADE = G_SHADE,
	GM_GOURAUD = G_SHADING_SMOOTH,
} GeometryMode;

typedef enum {
	Start,
	SetVertexSource,
	SetIndexSource,
	LoadVertexBuffer,
	LoadIndexBuffer,
	DrawPrimitives,
	SetBuffer,
	ClearBuffers,
	SetClearColor,
	SetClearZValue,
	LoadTexture,
	Jump,
	End,
	SetTextureScale,
	LoadMatrix,
	LoadLight,
	Sync,
	JumpToTranslated,
	SetCombineMode,
	SetRenderMode,
	NormalizePerspective,
	SetVertexScale,
	SetLookAt,
	SetGeometryMode,
	ClearGeometryMode,
	PipeSync,
	SetViewport,
	SetMaterial,
} CommandName;

typedef struct {
	name_t		name;
} Command_Start;

typedef struct {
	name_t		name;
	const void*	data;
	uint8_t		format;
} Command_SetVertexSource;

typedef struct {
	name_t		name;
	const void*	data;
} Command_SetIndexSource;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nvts;
} Command_LoadVertexBuffer;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nids;
} Command_LoadIndexBuffer;

typedef struct {
	name_t		name;
	uint8_t		type;
	uint32_t	num;
} Command_DrawPrimitives;

typedef struct {
	name_t		name;
	void*		buf;
	uint16_t	width;
	uint16_t	height;
	uint8_t		buftype;
	uint8_t		format;
} Command_SetBuffer;

typedef struct {
	name_t		name;
	uint8_t		types;
} Command_ClearBuffers;

typedef struct {
	name_t		name;
	float		r, g, b, a;
} Command_SetClearColor;

typedef struct {
	name_t		name;
	float		val;
} Command_SetClearZValue;

typedef struct {
	name_t		name;
	const Texture*	texture;
} Command_LoadTexture;

typedef struct {
	name_t		name;
	const DList*	dl;
} Command_Jump;

typedef struct {
	name_t		name;
} Command_End;

typedef struct {
	name_t		name;
	float		s;
	float		t;
} Command_SetTextureScale;

typedef struct {
	name_t		name;
	Matrix*		mtx;		// const Matrix* would be better, but libultra gets a pointer to a mutable value. Not sure how unsafe dropping const is.
	MatrixType	type;
} Command_LoadMatrix;

typedef struct {
	name_t		name;
	const Lux*	light;
	const Vector3f*	pos;
	uint16_t	idx;
} Command_LoadLight;

typedef struct {
	name_t		name;
} Command_Sync;

typedef struct {
	name_t		name;
} Command_PipeSync;

typedef struct {
	name_t		name;
	const void*	dl;
} Command_JumpToTranslated;

typedef struct {
	name_t		name;
	CombineMode	mode;
} Command_SetCombineMode;

typedef struct {
	name_t		name;
	RenderMode	mode;
} Command_SetRenderMode;

typedef struct {
	name_t		name;
	float		scale;
} Command_NormalizePerspective;

typedef struct {
	name_t		name;
	float		scale;
} Command_SetVertexScale;

typedef struct {
	name_t		name;
	LookAt*		lookat;
} Command_SetLookAt;

typedef struct {
	name_t		name;
	GeometryMode	gmode;
} Command_SetGeometryMode;

typedef struct {
	name_t		name;
	GeometryMode	gmode;
} Command_ClearGeometryMode;

typedef struct {
	name_t		name;
	const Viewport*	vp;
} Command_SetViewport;

typedef struct {
	name_t		name;
	const Material*	mtl;
} Command_SetMaterial;

void init_DList( DList* self );
void destroy_DList( DList* self );
void reset_DList( DList* self );
void* getData_DList( const DList* dlist );
size_t getSize_DList( const DList* dlist );

void pushStart_DList( DList* dlist );
void pushSetVertexSource_DList( DList* dlist, const void* vts, uint8_t format );
void pushSetIndexSource_DList( DList* dlist, const void* ids );
void pushLoadVertexBuffer_DList( DList* dlist, uint32_t src_ofs,
	uint32_t dst_ofs, uint32_t nvts );
void pushLoadIndexBuffer_DList( DList* dlist, uint32_t src_ofs,
	uint32_t dst_ofs, uint32_t nids );
void pushDrawPrimitives_DList( DList* dlist, uint8_t type, uint32_t num );
void pushSetBuffer_DList( DList* dlist, uint8_t buftype, uint8_t format,
	uint16_t width, uint16_t height, void* buf );
void pushClearBuffers_DList( DList* dlist, uint8_t types );
void pushSetClearColor_DList( DList* dlist, float r, float g, float b, float a );
void pushSetClearZValue_DList( DList* dlist, float val );
void pushLoadTexture_DList( DList* dlist, const Texture* texture );
void pushJump_DList( DList* dlist, const DList* dl );
void pushEnd_DList( DList* dlist );
void pushSetTextureScale_DList( DList* dlist, float s, float t);
void pushLoadMatrix_DList( DList* dlist, Matrix* mtx, MatrixType type );
void pushLoadLight_DList( DList* dlist, const Lux* light,
	const Vector3f* position, uint16_t idx );
void pushSync_DList( DList* dlist );
void pushPipeSync_DList( DList* dlist );
void pushJumpToTranslated_DList( DList* dlist, const void* dl );
void pushSetCombineMode_DList( DList* dlist, CombineMode mode );
void pushSetRenderMode_DList( DList* dlist, RenderMode mode );
void pushNormalizePerspective_DList( DList* dlist, float scale );
void pushSetVertexScale_DList( DList* dlist, float scale );
void pushSetLookAt_DList( DList* dlist, LookAt* lookat );
void pushSetGeometryMode_DList( DList* dlist, GeometryMode gmode );
void pushClearGeometryMode_DList( DList* dlist, GeometryMode gmode );
void pushSetViewport_DList( DList* dlist, const Viewport* vp );
void pushSetMaterial_DList( DList* dlist, const Material* material );

#endif	/* DLIST_H */

