#include <abort.h>
#include <malloc.h>
#include <ultra64.h>
#include <string.h>
#include <dltrans.h>
#include "sys.h"

#define STACKSIZE	0x2000
#define	NUM_PI_MSGS	8			/*  */
#define	MAINTHREAD_ID	3			/*  */
#define	MAINTHREAD_PRI	10			/*  */
#define	VBITHREAD_ID	4			/*  */
#define	VBITHREAD_PRI	12			/*  */
#define	SITHREAD_ID	5			/*  */
#define	SITHREAD_PRI	11			/*  */
#define AITHREAD_ID	6
#define AITHREAD_PRI	13

#define AUDIO_FREQUENCY	22050	//44100

#define	FRAMEBUFFER0			((void*)0x80200000)
#define	FRAMEBUFFER1			(void*)((uintptr_t)FRAMEBUFFER0 + BUFFER_W*BUFFER_H*4)
#define	AUXBUFFER			(void*)((uintptr_t)FRAMEBUFFER1 + BUFFER_W*BUFFER_H*4)
#define AUDIOBUFFER			(void*)((uintptr_t)AUXBUFFER + BUFFER_W*BUFFER_H*4)
#define AUDIOBUFFER_SAMPLES_CHNL	(735*2)
#define AUDIOBUFFER_SIZE		(AUDIOBUFFER_SAMPLES_CHNL*2*2)
#define ZBUFFER				((void*)0x80300000)

extern uint32_t __stack_adr;
extern uint32_t __bss_start, __bss_end;
extern void mainproc(void*);

static void idle(void*);

/* structures */
static struct SysState {
	/* thread stuff */
	OSThread	idleThread;
	OSThread	mainThread;
	OSThread	vbiThread;
	OSThread	siThread;
	OSThread	aiThread;
	u64		idleThreadStack[STACKSIZE/sizeof(u64)];
	u64		mainThreadStack[STACKSIZE/sizeof(u64)];
	u64		vbiThreadStack[STACKSIZE/sizeof(u64)];
	u64		siThreadStack[STACKSIZE/sizeof(u64)];
	u64		aiThreadStack[STACKSIZE/sizeof(u64)];

	/* various */
	OSPiHandle*	piHandle;
	OSTask		tlist;
	bool		isRendering;
	bool		aiThreadCreated;
	int		frameCnt;
	DLTrans		dltrans;
	NativeDL	displayList;
	GfxMode		gfxMode;
	const void*	audioSrc;
	uint32_t	aiFrequency;

	/* messages and message queues */
	OSMesgQueue	rspMessageQ;
	OSMesgQueue	rdpMessageQ;
	OSMesgQueue	viMessageQ;
	OSMesgQueue	vbiMessageQ;
	OSMesgQueue	siMessageQ;
	OSMesgQueue	aiMessageQ;
	OSMesg		rspMessages;
	OSMesg		rdpMessages;
	OSMesg		viMessages;
	OSMesg		vbiMessages;
	OSMesg		siMessages;
	OSMesg		aiMessages;

	/* surfaces */
	Surface*	colorBuffer;		/* surface to DRAW on */
	Surface		colorBuffer0;
	Surface		colorBuffer1;
	Surface		auxBuffer;
	Surface		zBuffer;

	/* handlers */
	void		(*handlerVbi)(void*);
	void		(*handlerController)(void*);

	/* RSP stuff. We use KSEG0 virtual adressing, assuming that RSP maps segment 8 to physical address 0x0 */
	u64		dram_stack[SP_DRAM_STACK_SIZE64] __attribute__((aligned (16)));
	u64		yieldBuffer[OS_YIELD_DATA_SIZE/sizeof(u64)];
} sysState;

void boot( void ) {
	for(uint32_t* bss = (uint32_t*)__bss_start; bss < (uint32_t*)__bss_end; bss++) {
		*bss = 0;
	}

	osInitialize();
	sysState.piHandle = osCartRomInit();
	osCreateThread(
		&sysState.idleThread,
		1,
		idle,
		(void*)0,
		sysState.idleThreadStack + STACKSIZE/sizeof(u64),
		10
	);
	osStartThread( &sysState.idleThread );
}

static void mainproc_wrapper( void* arg ) {
	init_Sys();

	mainproc( arg );

	raiseError_Sys();
}

static void aiThread( void* arg ) {
	if( arg ) {}

	osCreateMesgQueue(
		&sysState.aiMessageQ,
		&sysState.aiMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_AI,
		&sysState.aiMessageQ,
		NULL
	);
	osAiSetFrequency( sysState.aiFrequency );

	const uint8_t* src = sysState.audioSrc;
	uint8_t* dst = (uint8_t*)AUDIOBUFFER;
	for(uint32_t i=0; i<AUDIOBUFFER_SIZE; i+=2) {
		*dst++ = *(src+1);
		*dst++ = *(src);
		src += 2;
	}

	osAiSetNextBuffer(
		(uint8_t*)AUDIOBUFFER,
		AUDIOBUFFER_SIZE
	);

	for( int k=0; true; ) {
		//while(!(osAiGetStatus() & AI_STATUS_DMA_BUSY)) {}	// seems like PJ64 can't handle it properly
		osAiSetNextBuffer(
			(uint8_t*)AUDIOBUFFER
				+ (k ? 0 : AUDIOBUFFER_SIZE), AUDIOBUFFER_SIZE
		);

		/* wait till buffer is consumed */
		osRecvMesg(
			&sysState.aiMessageQ,
			NULL,
			OS_MESG_BLOCK
		);

		dst = (uint8_t*)AUDIOBUFFER + (k ? AUDIOBUFFER_SIZE : 0);
		for( uint32_t i=0; i<AUDIOBUFFER_SIZE; i+=2 ) {
			*dst++ = *(src+1);
			*dst++ = *(src);
			src += 2;
		}

		if( (uintptr_t)src >= (uintptr_t)sysState.audioSrc + AUDIOBUFFER_SIZE*2*14 ) {
			src = sysState.audioSrc;
		}
			
		osWritebackDCacheAll();

		k ^= 1;
	}
}

static void siThread( void* arg ) {
	if(arg) {}

	OSContPad rdata[ MAXCONTROLLERS ];

	while( 1 ) {
		osRecvMesg(
			&sysState.siMessageQ,
			NULL,
			OS_MESG_BLOCK
		);

		osContGetReadData( rdata );

		if( sysState.handlerController ) {
			sysState.handlerController( &rdata[0] );
		}
	}
}

static void vbiThread( void* arg ) {
	if(arg) {}

	osCreateMesgQueue(
		&sysState.vbiMessageQ,
		&sysState.vbiMessages,
		1
	);
	osViSetEvent(
		&sysState.vbiMessageQ,
		NULL,
		1
	);

	/* wait for vertical blank */
	while( 1 ) {
		osContStartReadData( &sysState.siMessageQ );
		if( MQ_IS_FULL( &sysState.vbiMessageQ ) ) {
			osRecvMesg(
				&sysState.vbiMessageQ,
				NULL,
				OS_MESG_BLOCK
			);
		}
		osRecvMesg(
			&sysState.vbiMessageQ,
			NULL,
			OS_MESG_BLOCK
		);
		sysState.frameCnt++;

		/* handle user */
		if( sysState.handlerVbi ) {
			sysState.handlerVbi( NULL );
		}
		/* forward notification to another queue */
		osJamMesg(
			&sysState.viMessageQ,
			NULL,
			NULL
		);
	}
}

static void idle(void* arg) {
	if(arg) {}

	osCreateViManager(
		OS_PRIORITY_VIMGR
	);
	osViSetMode(
		&osViModePalLpn1
	);
	osViSetSpecialFeatures(
		OS_VI_GAMMA_DITHER_OFF
	);
	//osCreatePiManager(OS_PRIORITY_PIMGR, &PiMessageQ, PiMessages, NUM_PI_MSGS);

	/* main thread */
	osCreateThread(
		&sysState.mainThread,
		MAINTHREAD_ID,
		mainproc_wrapper,
		(void*)0,
		sysState.mainThreadStack + STACKSIZE/sizeof(u64),
		MAINTHREAD_PRI
	);
	osStartThread( &sysState.mainThread );

	osSetThreadPri( 0, OS_PRIORITY_IDLE );

	while( 1 ) {
	}
}

int init_Sys( void ) {
	sysState.frameCnt = 0;
	sysState.isRendering = false;
	sysState.aiThreadCreated = false;

	/* Create message queue for VI; interestingly, a primitive drawn blinks without it */
	osCreateMesgQueue(
		&sysState.viMessageQ,
		&sysState.viMessages,
		1
	);
	//osViSetEvent(&sysState.viMessageQ, NULL, 1);

	/* Message queues for RSP/RDP */
	osCreateMesgQueue(
		&sysState.rspMessageQ,
		&sysState.rspMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_SP,
		&sysState.rspMessageQ,
		NULL
	);
	osCreateMesgQueue(
		&sysState.rdpMessageQ,
		&sysState.rdpMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_DP,
		&sysState.rdpMessageQ,
		NULL
	);

	/* Surfaces */
	init_Surface(
		&sysState.colorBuffer0,
		PIXELFMT_RGBA_5551,
		FRAMEBUFFER0,
		BUFFER_W,
		BUFFER_H
	);
	init_Surface(
		&sysState.colorBuffer1,
		PIXELFMT_RGBA_5551,
		FRAMEBUFFER1,
		BUFFER_W,
		BUFFER_H
	);
	init_Surface(
		&sysState.auxBuffer,
		PIXELFMT_RGBA_5551,
		AUXBUFFER,
		BUFFER_W,
		BUFFER_H
	);

	init_Surface(
		&sysState.zBuffer,
		PIXELFMT_16,
		ZBUFFER,
		BUFFER_W,
		BUFFER_H
	);

	sysState.colorBuffer = &sysState.colorBuffer0;

	/* Heap */
	void* heapstart = (void*)(
		(uintptr_t)getData_Surface(
			getBuffer_Sys( BUFFER_Z )
		)
		+
		getSize_Surface(
			getBuffer_Sys( BUFFER_Z )
		)
	);

	mallocInit(
		heapstart,
		__stack_adr - STACKSIZE - (u32)heapstart
	);

	/* other stuff */
	setAudioFrequency_Sys( AUDIO_FREQUENCY );
	init_NativeDL( &sysState.displayList );
	init_DLTrans( &sysState.dltrans );

	/* SI stuff */
	osCreateMesgQueue(
		&sysState.siMessageQ,
		&sysState.siMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_SI,
		&sysState.siMessageQ,
		NULL
	);

	u8 pattern;
	OSContStatus sdata[MAXCONTROLLERS];
	osContInit(
		&sysState.siMessageQ,
		&pattern,
		sdata
	);

	/* SI thread */
	sysState.handlerController = NULL;
	osCreateThread(
		&sysState.siThread,
		SITHREAD_ID,
		siThread,
		(void*)0,
		sysState.siThreadStack + STACKSIZE/sizeof(u64),
		SITHREAD_PRI
	);
	osStartThread( &sysState.siThread );

	/* VBI thread */
	sysState.handlerVbi = NULL;
	osCreateThread(
		&sysState.vbiThread,
		VBITHREAD_ID,
		vbiThread,
		(void*)0,
		sysState.vbiThreadStack + STACKSIZE/sizeof(u64),
		VBITHREAD_PRI
	);
	osStartThread( &sysState.vbiThread );

	return 0;
}

void waitForRSP_Sys( void ) {
	osRecvMesg(
		&sysState.rspMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

void waitForRDP_Sys( void ) {
	if( !isRendering_Sys() ) {
		return;
	}

	osRecvMesg(
		&sysState.rdpMessageQ,
		NULL,
		OS_MESG_BLOCK
	);

	sysState.isRendering = false;
}

inline void waitForGPU_Sys( void ) {
	waitForRDP_Sys();
}

void waitForVBI_Sys( void ) {
	if( MQ_IS_FULL( &sysState.viMessageQ ) ) {
		osRecvMesg(
			&sysState.viMessageQ,
			NULL,
			OS_MESG_BLOCK
		);
	}
	osRecvMesg(
		&sysState.viMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

void runGPU_Sys( const DList* dlist ) {
	resetCounter_NativeDL( &sysState.displayList );
	process_DLTrans(
		&sysState.dltrans,
		dlist,
		&sysState.displayList
	);

	OSTask_t* p = &sysState.tlist.t;
	p->type = M_GFXTASK;
	p->flags = OS_TASK_DP_WAIT;
	p->ucode_size = SP_UCODE_SIZE;
	p->ucode_data_size = SP_UCODE_DATA_SIZE;
	p->dram_stack = &sysState.dram_stack[0];
	p->dram_stack_size = SP_DRAM_STACK_SIZE8;
	p->ucode = (u64*)gspF3DEX2_xbusTextStart;
	p->ucode_data = (u64*)gspF3DEX2_xbusDataStart;
	p->ucode_boot = (u64*)rspbootTextStart;
	p->ucode_boot_size = (u32)rspbootTextEnd - (u32)rspbootTextStart;
	p->data_ptr = (u64*)getHead_NativeDL( &sysState.displayList );
	p->data_size = getLength_NativeDL( &sysState.displayList )*sizeof(Gfx);
	p->output_buff = NULL;
	p->output_buff_size = 0;
	p->yield_data_ptr = sysState.yieldBuffer;
	p->yield_data_size = OS_YIELD_DATA_SIZE;

	osWritebackDCacheAll();
	sysState.isRendering = true;
	osSpTaskStart( &sysState.tlist );
}

void swapBuffers_Sys( void ) {
	osViSwapBuffer(
		getData_Surface( sysState.colorBuffer )
	);

	waitForVBI_Sys();

	/* change surface pointer */
	if( sysState.colorBuffer == &sysState.colorBuffer0 ) {
		sysState.colorBuffer = &sysState.colorBuffer1;
	} else {
		sysState.colorBuffer = &sysState.colorBuffer0;
	}
}

const Surface* getBuffer_Sys( BufferType type ) {
	switch( type ) {
		case BUFFER_COLOR_DRAW:
			return sysState.colorBuffer;
		case BUFFER_Z:
			return &sysState.zBuffer;
		case BUFFER_AUX:
			return &sysState.auxBuffer;
		default:
			return NULL;
	}
}

inline NativeDL* getNativeDL_Sys( void ) {
	return &sysState.displayList;
}

void raiseError_Sys( void ) {
	for( uint32_t i=0; i<(BUFFER_W*BUFFER_H); ++i) {
		((uint8_t*)FRAMEBUFFER0)[i] = 0xFF;
		((uint8_t*)FRAMEBUFFER1)[i] = 0xFF;
	}
}

void registerHandler_Sys( HandlerType type, void (*handler)(void*) ) {
	switch( type ) {
		case HANDLER_VBI:
			sysState.handlerVbi = handler;
			break;
		case HANDLER_CONTROLLER:
			sysState.handlerController = handler;
			break;
		default:
			break;
	}
}

inline bool isRendering_Sys( void ) {
	return sysState.isRendering;
}

inline void* getHiMem_Sys( void ) {
	return (void*)(__stack_adr - STACKSIZE);
}

Time getTime_Sys( void ) {
	Time time;
	time.msec = sysState.frameCnt*(1000.0f/60);
	time.sec = time.msec / 1000;
	time.msec %= 1000;

	return time;
}

void setAudioSource_Sys( const void* src ) {
	if( sysState.aiThreadCreated ) {
		osDestroyThread( &sysState.aiThread );
	}

	sysState.audioSrc = src;

	osCreateThread(
		&sysState.aiThread,
		AITHREAD_ID,
		aiThread,
		(void*)0,
		sysState.aiThreadStack + STACKSIZE/sizeof(u64),
		AITHREAD_PRI
	);
	osStartThread( &sysState.aiThread );

	sysState.aiThreadCreated = true;
}

inline void setAudioFrequency_Sys( uint32_t frequency ) {
	sysState.aiFrequency = frequency;
}

